/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#ifndef __NV_FREEBSD_H__
#define __NV_FREEBSD_H__

#ifdef TRUE
#undef TRUE
#endif

#ifdef FALSE
#undef FALSE
#endif

#include <sys/param.h>

/*
 * This driver relies on a few changes to the FreeBSD kernel that are not
 * present in FreeBSD 4.6 or earlier - FreeBSD 5.0/-CURRENT is undergoing
 * active development and also unsupported.
 */

#if __FreeBSD_version < 470000
#error This driver requires FreeBSD 4.7 or later!
#endif

#include <sys/systm.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/pciio.h>
#include <sys/vnode.h>

#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/ioccom.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/sysent.h>
#include <sys/sysctl.h>

#include <machine/resource.h>
#include <machine/clock.h>
#include <machine/stdarg.h>
#include <machine/bus.h>
#include <machine/bus_memio.h>
#include <machine/vm86.h>

#include <sys/conf.h>
#include <sys/rman.h>
#include <sys/proc.h>
#include <sys/lock.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <sys/poll.h>

#include <sys/syscall.h>
#include <sys/bus.h>
#include <sys/memrange.h>
#include <sys/sysproto.h>
#include <sys/signalvar.h>

#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/vm_kern.h>
#include <vm/vm_page.h>
#include <vm/vm_extern.h>
#include <vm/vm_object.h>
#include <vm/pmap.h>
#include <vm/vm_map.h>

#include <pci/agpvar.h>
#include <sys/agpio.h>

#include "net/if.h"
#include "machine/../linux/linux.h"
#include "compat/linux/linux_ioctl.h"

#if __FreeBSD_version >= 500000
#include <sys/mutex.h>
#include <sys/smp.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#include <sys/filedesc.h>

/*
 * The resource manager client tracking needs an identifier that uniquely
 * represents a client connection across threads. It needs to be specific
 * to the thread that allocated the client. The Linux struct file pointer
 * is such an identifier. The per-process file descriptor table is enough
 * on FreeBSD, due to its thread-aware reference counting.
 */

#define __TD_FDT(td)     ((td)->td_proc->p_fd)
#define __TD_FDT_CNT(td) ((td)->td_proc->p_fd->fd_refcnt)

#else
#include <machine/smp.h>
#include <pci/pcireg.h>
#include <pci/pcivar.h>

#define vm_page_lock_queues()
#define vm_page_unlock_queues()

/*
 * The resource manager client tracking needs an identifier that uniquely
 * represents a client connection across threads. It needs to be specific
 * to the thread that allocated the client. The Linux struct file pointer
 * is such an identifier. The per-process file descriptor table is enough
 * on FreeBSD, due to its thread-aware reference counting.
 */

#define __TD_FDT(td)     ((td)->p_fd)
#define __TD_FDT_CNT(td) ((td)->p_fd->fd_refcnt)

#endif

/*
 * The NVIDIA kernel module's malloc identifier, needed for both tracking
 * and actual allocation/freeing purposes - declared here, but defined in
 * nvidia_os.c.
 */

MALLOC_DECLARE(M_NVIDIA);

/*
 * This define controls if the module will be built to support the NVIDIA
 * AGP GART driver or the FreeBSD AGPGART driver module. If you decide to
 * use AGPGART, the agp.ko module must be preloaded from loader.conf.
 */

#undef USE_OS_AGP_GART

/*
 * This flag determines if system memory for use with the NVIDIA AGP GART
 * driver is allocated as a contiguous block of kernel virtual memory, or
 * as a VM object. The latter is better (tm), and thus the default.
 */

#define NV_AGP_USE_VM_OBJECT 1

/*
 * This option decides if the driver will be built with support for Linux
 * compatibility. This makes nvidia.ko dependant on linux.ko; if you have
 * no need for Linux 3D applications, you can safely unset this flag.
 */

#define NV_SUPPORT_LINUX_COMPAT 1

/*
 * The DMA memory allocation tracking structure. DMA memory alloctions on
 * FreeBSD are tracked with their base address and size. Since PCI memory
 * is allocated from kernel virtual memory and since AGP allocations have
 * a contiguous range of addresses in the AGP aperture, base and size are
 * sufficient to track allocations.
 */

typedef
struct nv_alloc {
    SLIST_ENTRY(nv_alloc) list;
    u_int32_t size;
    vm_offset_t address;
    u_int32_t offset;
    struct vm_object *object;
} nv_alloc_t;

typedef
struct nv_alloc_private {
    u_int32_t count;
    u_int32_t alloc_type_contiguous;
    u_int32_t alloc_type_cached;
    u_int32_t alloc_type_kernel;
    u_int32_t class;
} nv_alloc_private_t;

typedef
struct nvidia_softc {
    device_t dev;
    device_t agp_dev;

    struct resource *reg;
    int reg_rid;
    int reg_type;

    struct resource *mem;
    int mem_rid;
    int mem_type;

    struct resource *irq;
    void *irq_ih;
    int   irq_rid;

    dev_t cdev;
    nv_state_t *nv_state;

    struct sysctl_ctx_list sysctl_ctx;
    struct selinfo rsel;

    struct callout_handle timer_ch;

    /* list of allocations */
    SLIST_HEAD(alloc_list, nv_alloc) alloc_list;

#if __FreeBSD_version >= 500000
    struct mtx mtx_rm;
    struct mtx mtx_api;
#else
    int spl;
    struct lock api_lock;
#endif
} nvidia_softc_t;


#define CDEV_MAJOR      180
#define CDEV_CTL_MINOR  255

extern devclass_t nvidia_devclass;
extern nv_state_t nvidia_ctl_state;

extern const char *pNVRM_ID;

#define PCIR_CAP_LIST_ID   0x00
#define PCIR_CAP_LIST_NEXT 0x01
#define PCIR_CAP_ID_AGP    0x02

/*
 * Entries in the NVIDIA glue-layer registry are now described by the new
 * shared nv_parm_t structure; please review nvidia_os_registry.c in case
 * you need to make low-level configuration changes. The entries are also
 * mapped into the SYSCTL hierarchy and thus easily accessible.
 */

extern nv_parm_t nv_parms[];


/* nvidia_dev.c */
int    nvidia_dev_attach     (struct nvidia_softc *);
int    nvidia_dev_detach     (struct nvidia_softc *);

/* nvidia_ctl.c */
int    nvidia_ctl_attach     (void);
int    nvidia_ctl_detach     (void);

/* nvidia_subr.c */
int    nvidia_attach         (device_t);
int    nvidia_detach         (device_t);
int    nvidia_suspend        (device_t);
int    nvidia_resume         (device_t);
int    nvidia_alloc          (device_t);
int    nvidia_free           (device_t);
void   nvidia_intr           (void *);
int    nvidia_modevent       (module_t, int, void *);

void   nvidia_rc_timer       (void *);

void   nv_lock_api           (nv_state_t *);
void   nv_unlock_api         (nv_state_t *);

S032   nv_alloc_contig_pages (nv_state_t *, VOID **, U032);
S032   nv_free_contig_pages  (nv_state_t *, VOID *);
S032   nv_alloc_system_pages (nv_state_t *, VOID **, U032);
S032   nv_free_system_pages  (nv_state_t *, VOID *);

S032   nv_alloc_vm_object    (nv_state_t *, VOID **, U032);
S032   nv_free_vm_object     (nv_state_t *, VOID *);

void*  nv_find_alloc         (nv_state_t *, vm_offset_t);
void*  nv_find_alloc_obj     (nv_state_t *, vm_offset_t);

S032   nv_os_agp_init        (nv_state_t *, VOID **, VOID **, U032 *);
S032   nv_os_agp_teardown    (nv_state_t *);
S032   nv_alloc_agp_pages    (nv_state_t *, VOID **, U032, U032, VOID **);
S032   nv_free_agp_pages     (nv_state_t *, VOID **, U032, VOID *);

/* nvidia_sysctl.c */
void   nvidia_sysctl_init    (void);
void   nvidia_sysctl_exit    (void);

U008   nvidia_find_cap       (device_t);
void*  nvidia_find_bridge    (void);

int    nvidia_sysctl_vbios   (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_type    (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_rates   (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_fw      (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_sba     (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_regs    (SYSCTL_HANDLER_ARGS);

int    nvidia_sysctl_driver  (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_rate_s  (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_fw_s    (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_sba_s   (SYSCTL_HANDLER_ARGS);
int    nvidia_sysctl_status  (SYSCTL_HANDLER_ARGS);

void   nv_sysctl_init        (nv_state_t *);
void   nv_sysctl_exit        (nv_state_t *);

/* nvidia_linux.c */
int    linux_ioctl_nvidia    (d_thread_t *, struct linux_ioctl_args *);
void   nvidia_linux_init     (void);
void   nvidia_linux_exit     (void);

/* ioctl helpers */
int    nvidia_set_primary     (struct nv_ioctl_primary_card *);
int    nvidia_get_card_info   (struct nv_ioctl_card_info *);
int    nvidia_get_api_version (struct nv_ioctl_rm_api_version *);
int    nvidia_handle_ioctl    (dev_t, u_long, caddr_t, int, d_thread_t *);


/* device helpers */
int    nvidia_open_ctl       (void);
int    nvidia_open_dev       (struct nvidia_softc *);
int    nvidia_close_ctl      (dev_t, d_thread_t *);
int    nvidia_close_dev      (struct nvidia_softc *, dev_t, d_thread_t *);
int    nvidia_mmap_dev       (struct nvidia_softc *, vm_offset_t, vm_offset_t *);

#endif /* __NV_FREEBSD_H__ */

