/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-freebsd.h"


devclass_t nvidia_devclass;
nv_state_t nvidia_ctl_state;

#define rman_get_extends(x) (rman_get_end(x) - rman_get_start(x) + 1)

int nvidia_attach(device_t dev)
{
    int status;
    struct nvidia_softc *sc;
    nv_state_t *nv;

    sc = device_get_softc(dev);
    nv = sc->nv_state;

    nv->os_state         = sc;
    nv->flags            = 0;
    nv->device_number    = device_get_unit(dev);
    nv->usage_count      = 0;
    nv->bus              = pci_get_bus(dev);
    nv->slot             = pci_get_slot(dev);
    nv->vendor_id        = pci_get_vendor(dev);
    nv->device_id        = pci_get_device(dev);
    nv->interrupt_line   = pci_get_irq(dev);
    nv->bar.fb.address   = rman_get_start(sc->mem);
    nv->bar.fb.size      = rman_get_extends(sc->mem);
    nv->bar.fb.map       = NULL;
    nv->bar.regs.address = rman_get_start(sc->reg);
    nv->bar.regs.size    = rman_get_extends(sc->reg);
    nv->bar.regs.map     = NULL;

    if ((status = nvidia_dev_attach(sc)) != 0)
        return status;

    if ((status = nvidia_ctl_attach()) != 0)
        return status;

    nv_sysctl_init(nv);
    return 0;
}

int nvidia_detach(device_t dev)
{
    int status;
    struct nvidia_softc *sc;

    sc = device_get_softc(dev);
    nv_sysctl_exit(sc->nv_state);

    if ((status = nvidia_dev_detach(sc)) != 0)
        return status;

    if ((status = nvidia_ctl_detach()) != 0)
        return status;

    return 0;
}


/*
 * We don't currently support power management on FreeBSD; these routines
 * are taken out until we do. Several steps are needed to support PM yet,
 * and quite a bit of testing.
 */

#if 0

#define LINUX_PM_SUSPEND    0
#define LINUX_PM_RESUME     1

int nvidia_suspend(device_t dev)
{
    struct nvidia_softc *sc;
    nv_state_t *nv;

    sc = device_get_softc(dev);
    nv = sc->nv_state;

    /* Only if ACPI is running */
    if (devclass_get_softc(devclass_find("ACPI"), 0))
        rm_power_management(nv, 0, LINUX_PM_SUSPEND);

    return 0;
}

int nvidia_resume(device_t dev)
{
    struct nvidia_softc *sc;
    nv_state_t *nv;

    sc = device_get_softc(dev);
    nv = sc->nv_state;

    /* Only if ACPI is running */
    if (devclass_get_softc(devclass_find("ACPI"), 0))
        RmPowerManagement(nv->pdev, 0, LINUX_PM_RESUME);

    return 0;
}

#endif /* power management */


int nvidia_alloc(device_t dev)
{
    int error = 0;
    struct nvidia_softc *sc;

    sc = device_get_softc(dev);
    sc->dev = dev;

    sc->nv_state = malloc(sizeof(nv_state_t), M_DEVBUF, M_WAITOK | M_ZERO);
    if (sc->nv_state == NULL) {
        error = ENOMEM;
        goto fail;
    }

    sc->reg = bus_alloc_resource(dev, sc->reg_type, &sc->reg_rid,
                    0, ~0, 1, 0);
    if (sc->reg == NULL) {
        device_printf(dev, "Unable to allocate NVIDIA register resource.\n");
        error = ENXIO;
        goto fail;
    }

    sc->mem = bus_alloc_resource(dev, sc->mem_type, &sc->mem_rid,
                    0, ~0, 1, 0);
    if (sc->mem == NULL) {
        device_printf(dev, "Unable to allocate NVIDIA memory resource.\n");
        error = ENXIO;
        goto fail;
    }

    sc->irq = bus_alloc_resource(dev, SYS_RES_IRQ, &sc->irq_rid,
                     0, ~0, 1, RF_SHAREABLE | RF_ACTIVE);
    if (sc->irq == NULL) {
        device_printf(dev, "Unable to allocate interrupt resource.\n");
        error = ENXIO;
        goto fail;
    }

fail:
    return (error);
}

int nvidia_free(device_t dev)
{
    struct nvidia_softc *sc;

    sc = device_get_softc(dev);

    if (sc->nv_state)
        free(sc->nv_state, M_NVIDIA); sc->nv_state = NULL;
    if (sc->reg)
        bus_release_resource(dev, sc->reg_type, sc->reg_rid, sc->reg);
    if (sc->mem)
        bus_release_resource(dev, sc->mem_type, sc->mem_rid, sc->mem);
    if (sc->irq_ih)
        bus_teardown_intr(dev, sc->irq, sc->irq_ih);
    if (sc->irq)
        bus_release_resource(dev, SYS_RES_IRQ, sc->irq_rid, sc->irq);

    return 0;
}

void nvidia_intr(void *xsc)
{
    struct nvidia_softc *sc;
    nv_state_t *nv;
    U032 run_bottom_half = 0;

    sc = (struct nvidia_softc *) xsc;
    nv = sc->nv_state;

    /*
     * The presence of both rm_isr and rm_isr_bh in this ISR looks rather
     * confusing at first, but makes sense once you know that it does not
     * run in a top-half context, but in a bottom-half one.
     */

    rm_isr(nv->device_number, &run_bottom_half);

    if (run_bottom_half) {
        rm_isr_bh(nv->pdev);
    }
}

int nvidia_set_primary(struct nv_ioctl_primary_card *pc)
{
    int i;
    struct nvidia_softc *sc;
    nv_state_t *nv;

    for (i = 0; i < NV_MAX_DEVICES; i++) {
        sc = devclass_get_softc(nvidia_devclass, i);
        if (!sc)
            continue;
        nv = sc->nv_state;

        if (nv->bus == pc->bus && nv->slot == pc->slot)
            nv->flags |= NV_FLAG_POSTED;
    }

    return 0;
}

int nvidia_get_api_version(struct nv_ioctl_rm_api_version *av)
{
    av->major   = NV_MAJOR_VERSION;
    av->minor   = NV_MINOR_VERSION;
    av->patch   = NV_PATCHLEVEL;
    av->version = NV_RM_API_VERSION;

    return 0;
}

int nvidia_get_card_info(struct nv_ioctl_card_info *ci)
{
    int i;
    struct nvidia_softc *sc;
    nv_state_t *nv;

    /*
     * Clients supporting versioning will pass version magic in the first
     * card information field.
     */
    struct nv_ioctl_rm_api_version *av = (void *) ci;

    if (av->magic != NV_RM_API_VERSION_MAGIC_REQ ||
        av->version != NV_RM_API_VERSION) {
        /*
         * This client either has no versioning support its version isn't
         * identical to ours. While this is fine most of the time, it can
         * result in severe problems.
         */
        av->magic   = NV_RM_API_VERSION_MAGIC_REP;
        av->major   = NV_MAJOR_VERSION;
        av->minor   = NV_MINOR_VERSION;
        av->patch   = NV_PATCHLEVEL;
        av->version = NV_RM_API_VERSION;

        return -EINVAL;
    }

    /* clear card information structure */
    memset(ci, 0, sizeof(ci));

    for (i = 0; i < NV_MAX_DEVICES; i++) {
        sc = devclass_get_softc(nvidia_devclass, i);
        if (!sc)
            continue;
        nv = sc->nv_state;

        ci[i].flags          = (NV_IOCTL_CARD_INFO_FLAG_PRESENT |
                                NV_IOCTL_CARD_INFO_FLAG_NEED_MSYNC);
        ci[i].bus            = nv->bus;
        ci[i].slot           = nv->slot;
        ci[i].vendor_id      = nv->vendor_id;
        ci[i].device_id      = nv->device_id;
        ci[i].interrupt_line = nv->interrupt_line;
        ci[i].fb_address     = nv->bar.fb.address;
        ci[i].fb_size        = nv->bar.fb.size;
        ci[i].reg_address    = nv->bar.regs.address;
        ci[i].reg_size       = nv->bar.regs.size;
    }

    return 0;
}


/*
 * The core resource manager's ioctl handler operates on the ioctl number
 * rather than the ioctl command, to allow it to be ignorant of operating
 * system specific encoding details. This macro extracts this number from
 * the incoming ioctl command.
 */

#define __IOC_NR(cmd) ((cmd) & 0xff)

int nvidia_handle_ioctl(
    dev_t dev,
    u_long cmd,
    caddr_t data,
    int fflag,
    d_thread_t *td
)
{
    struct nvidia_softc *sc;
    nv_state_t *nv;
    int unit = minor(dev);

    if (unit == CDEV_CTL_MINOR) {
        /* the control device is "special" */
        nv = &nvidia_ctl_state;
    } else {
        sc = devclass_get_softc(nvidia_devclass, unit);
        if (!sc)
            return ENXIO;
        nv = sc->nv_state;
    }

    if (rm_ioctl(nv, __TD_FDT(td), __IOC_NR(cmd), data))
        return 0;

    return EINVAL;
}


int nvidia_open_ctl(void)
{
    nv_state_t *nv = &nvidia_ctl_state;

    nv->device_number = NV_CONTROL_DEVICE_NUMBER;

    nv->flags |= NV_FLAG_OPEN;
    nv->flags |= NV_FLAG_CONTROL;

    nv->flags &= ~NV_FLAG_HOTKEY_OCCURRED;

    nv->usage_count++;
    return 0;
}

int nvidia_close_ctl(
    dev_t dev,
    d_thread_t *td
)
{
    nv_state_t *nv = &nvidia_ctl_state;

    if (__TD_FDT_CNT(td) == 0) {
        /*
         * The global open/close count allows us to determine when we can
         * safely reset this device; tracking of resource manager clients
         * is more involved. Hopefully, this solution works as expected.
         */
        rm_free_unused_clients(nv, 0, __TD_FDT(td));
    }

    nv->usage_count--;
    if (nv->usage_count == 0) {
        /*
         * The control device has been released; without physical devices
         * backing it, we only need to reset the flags.
         */
        nv->flags = 0;
    }

    return 0;
}

int nvidia_open_dev(struct nvidia_softc *sc)
{
    nv_state_t *nv = sc->nv_state;

    if ((nv->flags & NV_FLAG_OPEN) == 0) {
        /*
         * The device flags indicate that this device has not been opened
         * or else shut down previously. There isn't much we can do here,
         * the core resource manager does the actual work. We update both
         * the flags and the usage count.
         */
        if (!rm_init_adapter(nv)) {
            device_printf(sc->dev, "rm_init_adapter() failed!\n");
            return EIO;
        }

        nv->flags |= NV_FLAG_POSTED;
        nv->flags |= NV_FLAG_OPEN;
    }

    nv->usage_count++;
    return 0;
}

int nvidia_close_dev(
    struct nvidia_softc *sc,
    dev_t dev,
    d_thread_t *td
)
{
    nv_state_t *nv = sc->nv_state;

    if (__TD_FDT_CNT(td) == 0) {
        /*
         * The global open/close count allows us to determine when we can
         * safely reset this device; tracking of resource manager clients
         * is more involved. Hopefully, this solution works as expected.
         */
        rm_free_unused_clients(nv, 0, __TD_FDT(td));
    }

    nv->usage_count--;
    if (nv->usage_count == 0) {
        /*
         * The usage count for this device has dropped to zero, it can be
         * safely shut down. We don't need to wait for bottom-halfes like
         * we do on Linux, they are not run asynchronously on FreeBSD. We
         * do need to reset certain flags, though.
         */

        rm_disable_adapter(nv);
        rm_shutdown_adapter(nv);

        nv->flags &= ~(NV_FLAG_OPEN | NV_FLAG_WAITING);
    }

    return 0;
}


int nvidia_modevent(
    module_t mod,
    int what,
    void *arg
)
{
    switch (what) {
        case MOD_LOAD:
            /*
             * The module load event. Our KLD has just been loaded and is
             * ready to initialize. We setup the core resource manager in
             * this routine, further initialization takes place at attach
             * time.
             */
            if (!rm_init_rm()) {
                printf("nvidia: rm_init_rm() failed!\n");
                return EIO;
            }

            nvidia_sysctl_init();
            nvidia_linux_init();
            rm_load_registry(nv_parms);

            break;

        case MOD_UNLOAD:
            /*
             * The module unload event. At this point, all attached cards
             * have been shut down, it is safe to tear down core resman.
             */
            rm_shutdown_rm();

            nvidia_sysctl_exit();
            nvidia_linux_exit();

            break;

        default:
            break;
    }

    return 0;
}


#ifdef USE_OS_AGP_GART
S032 nv_os_agp_init(
    nv_state_t *nv,
    VOID **base,
    VOID **linear,
    U032 *limit
)
{
    void *bitmap;
    struct nvidia_softc *sc = nv->os_state;
    struct agp_info ai;

    U032 mode = 0;
    U032 fw   = 0;
    U032 sba  = 0;
    U032 rate = AGP_MODE_RATE_1x | AGP_MODE_RATE_2x | AGP_MODE_RATE_4x;
    U032 size = 0;

    sc->agp_dev = agp_find_device();
    if (!sc->agp_dev) {
        /*
         * Is the AGP device attached? Module dependency prevents us from
         * running without the AGP module loaded - but it may have failed
         * to attach.
         */
        printf("nvidia: agp_find_device failed, chipset unsupported?\n");
        return -ENODEV;
    }

    if (agp_acquire(sc->agp_dev) != 0) {
        /*
         * It seems that the AGP backend has already been claimed, or has
         * somehow entered a bad state (acquired, abandoned, not released
         * by a driver).
         */
        return -EBUSY;
    }

    agp_get_info(sc->agp_dev, &ai);
    mode = ai.ai_mode;

    if (os_set_mem_range(ai.ai_aperture_base, ai.ai_aperture_size,
                NV_MEMORY_WRITECOMBINED) != RM_OK) {
        /*
         * Failure to set a write-combining range for the AGP aperture is
         * not necessarily a fatal error condition; we don't know at this
         * point, however, and abort to prevent performance and stability
         * problems that may be hard to track down otherwise.
         */
        agp_release(sc->agp_dev);
    }

    rm_read_registry_dword(nv, "NVreg", "ReqAGPRate", &rate);
    rm_read_registry_dword(nv, "NVreg", "EnableAGPFW", &fw);
    rm_read_registry_dword(nv, "NVreg", "EnableAGPSBA", &sba);

    mode = AGP_MODE_SET_RATE(mode, mode & AGP_MODE_GET_RATE(mode));
    mode = AGP_MODE_SET_FW(mode, fw);
    mode = AGP_MODE_SET_SBA(mode, sba);

    if (agp_enable(sc->agp_dev, mode) != 0) {
        /*
         * At this point, most backends are enabled with generic code. It
         * does basic sanity checking on the mode we provide it with, but
         * shouldn't fail. This may change in the future.
         */
        agp_release(sc->agp_dev);
        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size);
        return -EIO;
    }

    size = ai.ai_aperture_size / RM_PAGE_SIZE / 8;
    
    if (os_alloc_mem(&bitmap, size) != RM_OK) {
        agp_release(sc->agp_dev);
        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size);
        return -EIO;
    }

    os_mem_set(bitmap, 0xff, size);

    if (rm_set_agp_bitmap(nv, bitmap) != RM_OK) {
        agp_release(sc->agp_dev);
        os_free_mem(bitmap);
        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size);
        return -EIO;
    }

    *base   = (VOID *) ai.ai_aperture_base;
    *linear = (VOID *) ai.ai_aperture_va;
    *limit  = (U032)   ai.ai_aperture_size - 1;

    return 0;
}

S032 nv_os_agp_teardown(nv_state_t *nv)
{
    struct nvidia_softc *sc = nv->os_state;
    void *bitmap;

    if (agp_release(sc->agp_dev) != 0) {
        /*
         * This can happen if we failed to free all allocated AGP buffers
         * prior to releasing the backend. This shouldn't happen, we have
         * code to prevent it.
         */
        return -EBUSY;
    }

    rm_clear_agp_bitmap(nv, &bitmap);
    os_free_mem(bitmap);

    os_unset_mem_range(nv->agp.address, nv->agp.size);

    return 0;
}
#endif /* USE_OS_AGP_GART */

S032 nv_agp_init(
    nv_state_t *nv,
    VOID **base,
    VOID **linear,
    VOID *limit,
    U032 config
)
{
    nv->agp_config = NVOS_AGP_CONFIG_DISABLE_AGP;

#ifdef USE_OS_AGP_GART
    if ((config & NVOS_AGP_CONFIG_OSAGP) == 0) {
        /*
         * If the user disabled AGP or requested that the NVIDIA built-in
         * AGP GART driver be used, we must abort; in the latter case, it
         * is necessary to rebuild the driver.
         */
        printf("nvidia: rebuild nvidia.ko for NVIDIA AGP GART support.\n");
        return 1;
    }

    if (nv_os_agp_init(nv, base, linear, limit) == 0) {
        /*
         * If the operating system AGP GART driver successfully setup the
         * backend, we need to make sure that we apply any overrides that
         * we deem necessary.
         */
        rm_update_agp_config(nv);
        nv->agp_config = NVOS_AGP_CONFIG_OSAGP;
        return 0;
    }

    return 1;
#else
    if ((config & NVOS_AGP_CONFIG_NVAGP) == 0) {
        /*
         * The user either disabled AGP (in which case we should not have
         * been called), or configured the driver to use the FreeBSD GART
         * driver. The latter case requires rebuilding the driver.
         */
        printf("nvidia: rebuild nvidia.ko for FreeBSD AGPGART support.\n");
        return 1;
    }

    if (rm_init_agp(nv) == RM_OK) {
        /*
         * The built-in AGP GART driver was successfully initialized. The
         * internal data structures have been updated; the state needs to
         * be updated to make the new configuration known.
         */
        nv->agp_config = NVOS_AGP_CONFIG_NVAGP;
        return 0;
    } else {
        /*
         * Initialization of the built-in AGP GART driver failed. Reasons
         * for this are most commonly unknown chipsets or lack of support
         * for MTRRs.
         */
        return 1;
    }
#endif /* USE_OS_AGP_GART */
}

S032 nv_agp_teardown(nv_state_t *nv)
{
    if (nv->agp_config == NVOS_AGP_CONFIG_DISABLE_AGP)
        return -EINVAL;

#ifdef USE_OS_AGP_GART
    if (nv_os_agp_teardown(nv) == 0) {
        /*
         * The AGP backend has been released, the AGP core logic is still
         * initialized, though. This is by design (FreeBSD AGP GART).
         */
        return 0;
    }
#else
    if (rm_teardown_agp(nv) == RM_OK) {
        /*
         * The built-in AGP GART driver shut down, all internal data have
         * been reset.
         */
        return 0;
    }
#endif /* USE_OS_AGP_GART */

    nv->agp_config = NVOS_AGP_CONFIG_DISABLE_AGP;

    return 1;
}


VOID* nv_find_agp_kernel_mapping(
    nv_state_t *nv,
    ULONG address
)
{
    /*
     * XXX This function is not actually needed on this platform, and can
     * be left unimplemented. Ideally, this stub wouldn't need to be here
     * in the first place.
     */
    return NULL;
}


void nv_lock_rm(nv_state_t *nv)
{
#if __FreeBSD_version >= 500000
    /*
     * With SMPng, the "giant" kernel lock is gone. That means that we're
     * in a more complex enviroment locking-wise, but since the necessary
     * locking primitives are available to us, we can handle it.
     *
     * With mtx_lock_spin we acquire a spin mutex and locally disable all
     * interrupts on the current processor.
     */
    struct nvidia_softc *sc = nv->os_state;
    mtx_lock_spin(&sc->mtx_rm);
#else
    /* 
     * While SMP configurations are handled with a global kernel lock and
     * in such a way that only one of the CPUs is executing a process and
     * servicing interrupts at any point in time, it is possible that the
     * interrupt handlers (th/bh) may interfere with critical code.
     *
     * This "lock" thus disables interrupts for the duration of this code
     * segment's execution.
     */
    struct nvidia_softc *sc = nv->os_state;
    sc->spl = splhigh();
#endif
}

void nv_unlock_rm(nv_state_t *nv)
{
#if __FreeBSD_version >= 500000
    /*
     * Release the spin mutex acquired earlier and reenable interrupts on
     * this processor.
     */
    struct nvidia_softc *sc = nv->os_state;
    mtx_unlock_spin(&sc->mtx_rm);
#else
    /*
     * Restore the interrupt mask upon existing the critical code segment
     * and thus allow interrupt handlers to execute again.
     */
    struct nvidia_softc *sc = nv->os_state;
    splx(sc->spl);
#endif
}


void nv_lock_api(nv_state_t *nv)
{
#if __FreeBSD_version >= 500000
    /*
     * Serialize all API calls using a context switch mutex; since we are
     * not protecting against the interrupt handler with this lock, there
     * is no need to spin/disable interrupts.
     */
    struct nvidia_softc *sc = nv->os_state;
    mtx_lock(&sc->mtx_api);
#else
    /*
     * Due to FreeBSD 4.x's architectural approach towards SMP setups, we
     * know that no two parts of our driver can execute concurrently with
     * the exception of ISRs and system calls.
     *
     * Since it is entirely possible that we involuntarily go to sleep at
     * certain points in time, we make sure that we execute only a single
     * of our system calls at a time.
     */
    struct nvidia_softc *sc = nv->os_state;
    lockmgr(&sc->api_lock, LK_EXCLUSIVE, 0, curproc);
#endif
}

void nv_unlock_api(nv_state_t *nv)
{
#if __FreeBSD_version >= 500000
    /*
     * Release the API mutex; if a higher priority thread was waiting for
     * this mutex, the current thread will be preempted.
     */
    struct nvidia_softc *sc = nv->os_state;
    mtx_unlock(&sc->mtx_api);
#else
    /*
     * Release the API lock and allow a process sleeping on it to execute
     * its system call.
     */
    struct nvidia_softc *sc = nv->os_state;
    lockmgr(&sc->api_lock, LK_RELEASE, 0, curproc);
#endif
}


void nv_post_event(void *filep)
{
    /*
     * XXX This function is not yet implemented, but is also not required
     * to guarantee proper function of this driver. The mechanism that it
     * is part of is likely to be replaced with a more flexible one.
     */
    return;
}

void nv_post_vblank(nv_state_t *nv)
{
    /*
     * XXX This function is not yet implemented, but is also not required
     * to guarantee proper function of this driver. The mechanism that it
     * is part of is likely to be replaced with a more flexible one.
     */
    return;
}

void* nv_find_alloc(
    nv_state_t *nv,
    vm_offset_t address
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;

    SLIST_FOREACH(at, &sc->alloc_list, list) {
        /*
         * This lookup is straight-forward; the caller expects to find an
         * allocation matching the incoming base address. This routine is
         * only used to find allocations that are to be freed.
         */
        if (at->address == address)
            return (void *) at;
    }

    return NULL;
}

void* nv_find_alloc_obj(
    nv_state_t *nv,
    vm_offset_t object
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;

    SLIST_FOREACH(at, &sc->alloc_list, list) {
        /*
         * Similar to nv_find_alloc, this lookup is simple; the caller is
         * looking for an allocation matching the incoming offset to free
         * it.
         */
        if (at->object == (vm_object_t) object)
            return (void *) at;
    }

    return NULL;
}

VOID* nv_find_kernel_mapping(
    nv_state_t *nv,
    ULONG address
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;
    vm_offset_t offset, linear;

    offset = (vm_offset_t) address & PAGE_MASK;
    address &= ~PAGE_MASK;

    SLIST_FOREACH(at, &sc->alloc_list, list) {
        if (at->address == 0) {
            /*
             * XXX  This is the case for vm_object allocations, which are
             * bound to AGP allocations; this should test type flags.
             */
            continue;
        }

        linear = at->address;
        do {
            if (vtophys(linear) == (vm_offset_t) address) {
                /*
                 * XXX Compare against every page in the allocation; this
                 * relies on a linear kernel mapping.
                 */
                return (void *)(linear | offset);
            }
            linear += PAGE_SIZE;
        } while (linear < (at->address + at->size));
    }

    return NULL;
}

S032 nv_agp_translate_address(
    nv_state_t *nv,
    VOID *base,
    U032 index,
    U032 *address
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;

#ifdef NV_AGP_USE_VM_OBJECT
    vm_page_t m;

    SLIST_FOREACH(at, &sc->alloc_list, list) {
        /*
         * The purpose of this function is to extract the actual physical
         * addresses of each page in the allocation with the base address
         * passed by the caller.
         */
        if (at->object == (vm_object_t) base &&
            at->size > index * PAGE_SIZE) {
            m = vm_page_lookup(at->object, index);
            *address = VM_PAGE_TO_PHYS(m);
            return 0;
        }
    }
#else
    SLIST_FOREACH(at, &sc->alloc_list, list) {
        /*
         * The purpose of this function is to extract the actual physical
         * addresses of each page in the allocation with the base address
         * passed by the caller.
         */
        if (at->address == (vm_offset_t) base &&
            at->size > index * PAGE_SIZE) {
            *address = vtophys(at->address + (index * PAGE_SIZE));
            return 0;
        }
    }
#endif /* NV_AGP_USE_VM_OBJECT */

    return -1;
}

S032 nv_alloc_contig_pages(
    nv_state_t *nv,
    VOID **address,
    U032 count
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;
    U032 size = count * PAGE_SIZE;

    /*
     * XXX I'm not sure if we need to wire these pages; I'll assume there
     * is no need to for now.
     */
    if (os_alloc_contig_pages(address, size) != RM_OK)
        return -ENOMEM;

    at = malloc(sizeof(nv_alloc_t), M_NVIDIA, M_WAITOK | M_ZERO);
    if (!at) {
        os_free_contig_pages(*address, size);
        return -ENOMEM;
    }

    at->object = 0;
    at->size = size;
    at->address = (vm_offset_t) *address;
    at->offset = ~0;

    SLIST_INSERT_HEAD(&sc->alloc_list, at, list);

    return 0;
}

S032 nv_free_contig_pages(
    nv_state_t *nv,
    VOID *address
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;
    U032 size;

    at = nv_find_alloc(nv, (vm_offset_t) address);
    if (!at) {
        /*
         * If this happens, something really bad must have happened; core
         * resman won't deliberately call us with bogus addresses.
         */
        os_dbg_breakpoint();
        return -EINVAL;
    }

    size = at->size;
    SLIST_REMOVE(&sc->alloc_list, at, nv_alloc, list);

    free(at, M_NVIDIA);
    os_free_contig_pages(address, size);

    return 0;
}

S032 nv_alloc_system_pages(
    nv_state_t  *nv,
    VOID **address,
    U032 count
)
{
    nv_alloc_t *at;
    vm_offset_t pa;
    struct nvidia_softc *sc = nv->os_state;

    void *vm;
    u_int32_t i, size = count * PAGE_SIZE;

    at = malloc(sizeof(nv_alloc_t), M_NVIDIA, M_WAITOK | M_ZERO);
    if (!at) {
        return -ENOMEM;
    }

    vm = malloc(size, M_NVIDIA, M_WAITOK | M_ZERO);
    if (!vm) {
        /*
         * This error condition is not unlikely to occur. We allocate our
         * DMA memory from kernel virtual memory to allow for our mmap(2)
         * implementation to work properly. FreeBSD has plenty of virtual
         * memory, but it is conceivable that we run out.
         */
        free(at, M_NVIDIA);
        return -ENOMEM;
    }

    for (i = 0; i < count; i++) {
        /*
         * Mark this page as wired down; this may prevent the kernel from
         * paging it out while we're still using it. This is necessary to
         * prevent random crashes.
         */
        pa = vtophys((vm_offset_t) vm + (i * PAGE_SIZE));
        vm_page_lock_queues();
        vm_page_wire(PHYS_TO_VM_PAGE(pa));
        vm_page_unlock_queues();
    }

    at->object = 0;
    at->size = size;
    at->address = (vm_offset_t) vm;
    at->offset = ~0;

    SLIST_INSERT_HEAD(&sc->alloc_list, at, list);
    *address = (void *) at->address;

    return 0;
}

S032 nv_free_system_pages(
    nv_state_t *nv,
    VOID *address
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;

    vm_offset_t pa;
    u_int32_t i, count;

    at = nv_find_alloc(nv, (vm_offset_t) address);
    if (!at) {
        /*
         * If this happens, something really bad must have happened; core
         * resman won't deliberately call us with bogus addresses.
         */
        os_dbg_breakpoint();
        return -EINVAL;
    }

    count = at->size / PAGE_SIZE;
    SLIST_REMOVE(&sc->alloc_list, at, nv_alloc, list);

    vm_page_lock_queues();
    for (i = 0; i < count; i++) {
        /*
         * Release "one wiring" of this page; this may reenable paging as
         * a result, depending on the total number of "wirings".
         */
        pa = vtophys(at->address + (i * PAGE_SIZE));
        vm_page_unwire(PHYS_TO_VM_PAGE(pa), 0);
    }
    vm_page_unlock_queues();

    free(at, M_NVIDIA);
    free(address, M_NVIDIA);

    return 0;
}


#ifdef NV_AGP_USE_VM_OBJECT
S032 nv_alloc_vm_object(
    nv_state_t *nv,
    VOID **address,
    U032 count
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;

    u_int32_t i;
    vm_page_t m;
    vm_object_t object;

    at = malloc(sizeof(nv_alloc_t), M_NVIDIA, M_WAITOK | M_ZERO);
    if (!at) {
        return -ENOMEM;
    }

    object = vm_object_allocate(OBJT_DEFAULT, count);

    for (i = 0; i < count; i++) {
        /*
         * Find the individual pages belonging to the allocation from the
         * VM object; since this is the first time they're accessed, they
         * will be allocated.
         */

#if __FreeBSD_version >= 500000
        /*
         * Request that this range of pages be wired down to prevent them
         * from being paged out. This is done below for FreeBSD 4.
         */
        m = vm_page_grab(object, i, VM_ALLOC_NORMAL | VM_ALLOC_WIRED);
#else
        m = vm_page_grab(object, i, VM_ALLOC_NORMAL);
#endif
        if (!m) {
            vm_page_lock_queues();
            for (i -= 1; i >= 0; i--) {
                m = vm_page_lookup(object, i);
                vm_page_wakeup(m);
                vm_page_unwire(m, 0);
            }
            vm_page_unlock_queues();
            return -ENOMEM;
        }

#if __FreeBSD_version < 500000
        /*
         * Mark this page as wired down; this may prevent the kernel from
         * paging it out while we're still using it. This is necessary to
         * prevent random crashes.
         */
        vm_page_wire(m);
#endif
        vm_page_lock_queues();
        vm_page_wakeup(m);
        vm_page_unlock_queues();
    }

    at->object = object;
    at->offset = ~0;
    at->size = ptoa(count);
    at->address = 0;

    SLIST_INSERT_HEAD(&sc->alloc_list, at, list);
    *address = (void *) at->object;

    return 0;
}

S032 nv_free_vm_object(
    nv_state_t *nv,
    VOID *address
)
{
    nv_alloc_t *at;
    struct nvidia_softc *sc = nv->os_state;

    vm_page_t m;
    u_int32_t i, count;

    at = nv_find_alloc_obj(nv, (vm_offset_t) address);
    if (!at) {
        /*
         * If this happens, something really bad must have happened; core
         * resman won't deliberately call us with bogus addresses.
         */
        os_dbg_breakpoint();
        return -EINVAL;
    }

    count = at->size / PAGE_SIZE;
    SLIST_REMOVE(&sc->alloc_list, at, nv_alloc, list);

    vm_page_lock_queues();
    for (i = 0; i < count; i++) {
        /*
         * Release "one wiring" of this page; this may reenable paging as
         * a result, depending on the total number of "wirings".
         */
        m = vm_page_lookup(at->object, i);
        vm_page_unwire(m, 0);
    }
    vm_page_unlock_queues();

    vm_object_deallocate(at->object);
    free(at, M_NVIDIA);

    return 0;
}
#endif /* NV_AGP_USE_VM_OBJECT */


#ifdef USE_OS_AGP_GART
S032 nv_alloc_agp_pages(
    nv_state_t *nv,
    VOID **address,
    U032 count,
    U032 offset,
    VOID **private
)
{
    void *handle;
    struct nvidia_softc *sc = nv->os_state;

    handle = agp_alloc_memory(sc->agp_dev, 0, count << PAGE_SHIFT);
    if (!handle) {
        /*
         * This is very unlikely to happen, the system's memory resources
         * would have to be nearly exhausted.
         */
        return -ENOMEM;
    }

    if (agp_bind_memory(sc->agp_dev, handle, offset) != 0) {
        /*
         * This shouldn't happen, we claimed the AGP backend and are thus
         * using it exclusively; the resource manager manages AGP offsets
         * internally, we wouldn't have been called had we run out of AGP
         * aperture space.
         */
        os_dbg_breakpoint();

        agp_free_memory(sc->agp_dev, handle);
        return -ENOMEM;
    }

    *private = handle;
    return 0;
}

S032 nv_free_agp_pages(
    nv_state_t *nv,
    VOID **address,
    U032 count,
    VOID *handle
)
{
    struct nvidia_softc *sc = nv->os_state;

    if (agp_unbind_memory(sc->agp_dev, handle) != 0) {
        /*
         * This is the only place where previously bound AGP memory would
         * be freed. If we fail to unbind this memory now, something very
         * wrong must have happened.
         */
        os_dbg_breakpoint();
    }

    agp_free_memory(sc->agp_dev, handle);
    return 0;
}
#endif /* USE_OS_AGP_GART */


S032 nv_alloc_pages(
    nv_state_t *nv,
    VOID **address,
    U032 count,
    U032 alloc_type_agp,
    U032 alloc_type_contiguous,
    U032 alloc_type_cached,
    U032 alloc_type_kernel,
    U032 class,
    VOID **private
)
{
    if (alloc_type_agp) {
#ifdef USE_OS_AGP_GART
        U032 offset;

        if ((nv->agp_config & NVOS_AGP_CONFIG_OSAGP) == 0)
            return -EINVAL;

        if (rm_alloc_agp_bitmap(nv, count, &offset) != RM_OK) {
            /*
             * We've run out of AGP aperture space, the requestor will be
             * forced to fall back to general system memory.
             */
            return -ENOMEM;
        }

        if (nv_alloc_agp_pages(nv, address, count, offset << PAGE_SHIFT,
                private) != 0) {
            /*
             * The AGP aperture is managed internally by core resman, the
             * offset passed to us should work; thus, this shouldn't fail
             * unless the system memory is exhausted.
             */
            rm_free_agp_bitmap(nv, offset, count);
            return -ENOMEM;
        }

        *address = (void *)(nv->agp.address + (offset << PAGE_SHIFT));
        return 0;
#else
        nv_alloc_t *at;

        if ((nv->agp_config & NVOS_AGP_CONFIG_NVAGP) == 0)
            return -EINVAL;

#ifdef NV_AGP_USE_VM_OBJECT
        if (nv_alloc_vm_object(nv, address, count) != 0) {
            /*
             * Hopefully, this error condition will be much more unlikely
             * to occur than failure to allocate the same number of pages
             * with nv_alloc_system_pages.
             */
            return -ENOMEM;
        }

        at = nv_find_alloc_obj(nv, (vm_offset_t) *address);
        if (!at) {
            /*
             * This would be most peculiar, considering we just allocated
             * count page of system memory with nv_alloc_vm_object, which
             * returned *address as the key (the object).
             */
            os_dbg_breakpoint();
            return -ENOMEM;
        }

        if (rm_alloc_agp_pages(nv, address, count, class, private,
                &at->offset) != 0) {
            /*
             * In theory, this should only fail if we run out of aperture
             * space (which is not unlikely).
             */
            nv_free_vm_object(nv, *address);
            return -ENOMEM;
        }
#else
        if (nv_alloc_system_pages(nv, address, count) != 0) {
            /*
             * XXX This could fail for the reasons given above. We do not
             * actually have to allocate the pages for AGP like we do for
             * PCI DMA memory - the mmap(2) requirements are satisfied by
             * the AGP address ranges in the AGP aperture.
             */
            return -ENOMEM;
        }

        at = nv_find_alloc(nv, (vm_offset_t) *address);
        if (!at) {
            /*
             * This would be most peculiar, considering we just allocated
             * several system memory pages with nv_alloc_system_pages and
             * were given *address as the key (base address).
             */
            os_dbg_breakpoint();
            return -ENOMEM;
        }

        if (rm_alloc_agp_pages(nv, address, count, class, private,
                &at->offset) != 0) {
            /*
             * In theory, this should only fail if we run out of aperture
             * space (which is not unlikely).
             */
            nv_free_system_pages(nv, *address);
            return -ENOMEM;
        }
#endif /* NV_AGP_USE_VM_OBJECT */

        *address = (void *)(nv->agp.address + (at->offset << PAGE_SHIFT));
        return 0;
#endif /* USE_OS_AGP_GART */
    } else {
        /*
         * Save the allocation parameters for use in nv_free_pages, where
         * we won't have access to all of them (sigh).
         */
        nv_alloc_private_t *ap;

        ap = malloc(sizeof(nv_alloc_private_t), M_NVIDIA, M_WAITOK | M_ZERO);
        if (!ap) {
            return -ENOMEM;
        }
        *private = (void *) ap;

        ap->count = count;
        ap->alloc_type_contiguous = alloc_type_contiguous;
        ap->alloc_type_cached = alloc_type_cached;
        ap->alloc_type_kernel = alloc_type_kernel;
        ap->class = class;

        if (alloc_type_kernel &&
            !alloc_type_contiguous && alloc_type_cached) {
            /*
             * Obtain general system memory for the resource mananger, it
             * doesn't need to be contiguous.
             */
            if (nv_alloc_system_pages(nv, address, count))
                goto failed;

        } else if (alloc_type_contiguous ||
                   (alloc_type_kernel && !alloc_type_cached)) {
            /*
             * Make sure the pages are contiguous in physical memory; the
             * memory will not be used for DMA push buffers, the returned
             * linear mapping is expected to be usable as-is.
             */
            if (alloc_type_kernel && !alloc_type_cached) {
                /*
                 * XXX  allocate memory and get an uncached kernel linear
                 * mapping to it.
                 */
                goto failed;
            }

            if (nv_alloc_contig_pages(nv, address, count))
                goto failed;

        } else if (!alloc_type_kernel) {
            /*
             * Allocate general system memory for DMA push buffers. These
             * pages may be non-contiguous; the individual addresses must
             * be retrievable from the mmap(2) implementation.
             */
            if (nv_alloc_system_pages(nv, address, count))
                goto failed;

            (vm_offset_t)(uintptr_t) *address -= VM_MIN_KERNEL_ADDRESS;
        }

        return 0;

failed:
        free(*private, M_NVIDIA);
        return -ENOMEM;
    }
}

S032 nv_free_pages(
    nv_state_t *nv,
    VOID **address,
    U032 count,
    U032 alloc_type_agp,
    VOID *private
)
{
    U032 offset;
    offset = ((U032) *address - nv->agp.address) >> PAGE_SHIFT;

    if (alloc_type_agp) {
#ifdef USE_OS_AGP_GART
        if ((nv->agp_config & NVOS_AGP_CONFIG_OSAGP) == 0)
            return -EINVAL;

        if (nv_free_agp_pages(nv, address, count, private) != 0) {
            /*
             * The following is a sanity check, this mustn't ever fail in
             * in real-world scenario.
             */
            os_dbg_breakpoint();
            return -EINVAL;
        }

        rm_free_agp_bitmap(nv, offset, count);
#else
        nv_alloc_t *at;
        struct nvidia_softc *sc = nv->os_state;

        if ((nv->agp_config & NVOS_AGP_CONFIG_NVAGP) == 0)
            return -EINVAL;

        if (rm_free_agp_pages(nv, address, private) != 0) {
            /*
             * Don't free the underlying system memory unless the request
             * to free the RM AGP resources completes successfully, which
             * should never actually fail.
             */
            os_dbg_breakpoint();
            return -EINVAL;
        }

#ifdef NV_AGP_USE_VM_OBJECT
        SLIST_FOREACH(at, &sc->alloc_list, list) {
            /*
             * At allocation time, the AGP aperture offset was associated
             * with the underlying system memory. Find this system memory
             * based on the offset corresponding to the incoming aperture
             * address and free it.
             */
            if (at->offset == offset) {
                nv_free_vm_object(nv, (void *)at->object);
                break;
            }
#else
        SLIST_FOREACH(at, &sc->alloc_list, list) {
            /*
             * At allocation time, the AGP aperture offset was associated
             * with the underlying system memory. Find this system memory
             * based on the offset corresponding to the incoming aperture
             * address and free it.
             */
            if (at->offset == offset) {
                nv_free_system_pages(nv, (void *)at->address);
                break;
            }
#endif
        }
#endif /* USE_OS_AGP_GART */

        return 0;
    } else {
        nv_alloc_private_t *ap = (nv_alloc_private_t *) private;
        if (!ap) {
            /*
             * I don't expect this to fail; if it does, something bad has
             * been done to this memory.
             */
            os_dbg_breakpoint();
            return -EINVAL;
        }

        if (ap->alloc_type_kernel &&
            !ap->alloc_type_contiguous && ap->alloc_type_cached) {
            /*
             * Generic system memory used by the resource manager; almost
             * identical to the user memory case.
             */
            if (nv_free_system_pages(nv, *address))
                return -EINVAL;

        } else if (ap->alloc_type_contiguous ||
                   (ap->alloc_type_kernel && !ap->alloc_type_cached)) {
            /*
             * Free contigous system memory; this resource is rarely used
             * and hence managed globally.
             */
            if (ap->alloc_type_kernel && !ap->alloc_type_cached) {
                /*
                 * XXX We don't currently allocate this kind of memory in
                 * nv_alloc_pages, fix this case when we do.
                 */
                return -EINVAL;
            }

            if (nv_free_contig_pages(nv, *address))
                return -EINVAL;

        } else if (!ap->alloc_type_kernel) {
            /*
             * XXX Generic system memory used for DMA push buffers. There
             * are munmap(2) considerations in this case on Linux, are we
             * really safe here?
             */
            (vm_offset_t)(uintptr_t) *address += VM_MIN_KERNEL_ADDRESS;

            if (nv_free_system_pages(nv, *address))
                return -EINVAL;
        }

        free(ap, M_NVIDIA);
        return 0;
    }

    return -EINVAL;
}


ULONG nv_get_phys_address(ULONG address)
{
    struct vmspace *vm;
    vm_offset_t va = (vm_offset_t) address;

    if (va >= VM_MIN_KERNEL_ADDRESS)
        return vtophys(va);

    if (vm_fault_quick((caddr_t) va, VM_PROT_WRITE))
        return 0;

    vm = curproc->p_vmspace;
    return pmap_extract(vmspace_pmap(vm), va);
}


int nvidia_mmap_dev(
    struct nvidia_softc *sc,
    vm_offset_t offset,
    vm_offset_t *physical
)
{
    nv_alloc_t *at;
    nv_state_t *nv = sc->nv_state;

    /*
     * Offsets that fall into the frame buffer, registry or AGP apertures
     * are physical addresses and mapped into user-space directly. We can
     * only do some basic sanity checking here.
     */
    if (IS_FB_OFFSET(nv, offset, PAGE_SIZE)) {
        *physical = offset;
        return 0;
    }
    
    if (IS_REG_OFFSET(nv, offset, PAGE_SIZE)) {
        *physical = offset;
        return 0;
    }
    
    if (IS_AGP_OFFSET(nv, offset, PAGE_SIZE)) {
        *physical = offset;
        return 0;
    }

    offset += VM_MIN_KERNEL_ADDRESS;

    /*
     * If the offset does not fall into any of the relevant apertures, we
     * must assume that it corresponds to a memory allocation. Offsets of
     * this kind are kernel virtual addresses that fall within one of the
     * allocated ranges, provided they are valid. These logical addresses
     * are translated into physical addresses and then mapped.
     */
    SLIST_FOREACH(at, &sc->alloc_list, list) {
        if (offset >= at->address &&
                offset < at->address + at->size) {
            *physical = vtophys(offset);
            return 0;
        }
    }

    return -1;
}


int nv_int10h_call(
    nv_state_t * nv,
    U032 *eax,
    U032 *ebx,
    U032 *ecx,
    U032 *edx,
    void *buffer
)
{
    struct vm86frame vmf;

    vmf.vmf_eax = *eax;
    vmf.vmf_ebx = *ebx;
    vmf.vmf_ecx = *ecx;
    vmf.vmf_edx = *edx;

    if (vm86_intcall(0x10, &vmf) == 0) {
        *eax = vmf.vmf_eax;
        *ebx = vmf.vmf_ebx;
        *ecx = vmf.vmf_ecx;
        *edx = vmf.vmf_edx;
    } else {
        if (*eax == 0x4f14 && *ebx == 0x0186) {
            *eax = 0x004f;
            *ecx = ((nv->regs->map[0x101000/4] >> 16) & 0xf);
        }
    }

    return 0;
}

void nvidia_rc_timer(void *data)
{
    nv_state_t *nv = (nv_state_t *) data;
    struct nvidia_softc *sc;
    
    rm_run_rc_callback(nv);
    /*
     * We need this timer to trigger again one second from now, reset the
     * timeout.
     */
    sc = nv->os_state;
    sc->timer_ch = timeout(nvidia_rc_timer, (void *) nv, hz);
}

int nv_start_rc_timer(
    nv_state_t *nv
)
{
    struct nvidia_softc *sc = nv->os_state;

    if (nv->rc_timer_enabled != 0)
        return -EIO;

    sc->timer_ch = timeout(nvidia_rc_timer, (void *) nv, hz);
    nv->rc_timer_enabled = 1;

    return 0;
}

int nv_stop_rc_timer(
    nv_state_t *nv
)
{
    struct nvidia_softc *sc = nv->os_state;

    if (nv->rc_timer_enabled == 0)
        return -EIO;

    untimeout(nvidia_rc_timer, (void *) nv, sc->timer_ch);
    nv->rc_timer_enabled = 0;

    return 0;
}

